@@ -457,6 +457,17 @@ class Irreps(tuple):
         return ir in (irrep for _, irrep in self)
 
     def count(self, ir) -> int:
+        r"""Multiplicity of ``ir``.
+
+        Parameters
+        ----------
+        ir : `Irrep`
+
+        Returns
+        -------
+        `int`
+            total multiplicity of ``ir``
+        """
         ir = Irrep(ir)
         return sum(mul for mul, irrep in self if ir == irrep)
 
@@ -554,6 +565,8 @@ class Irreps(tuple):
 
     @property
     def lmax(self) -> int:
+        if len(self) == 0:
+            raise ValueError("Cannot get lmax of empty Irreps")
         return max(self.ls)
 
     def __repr__(self):
