@@ -1,4 +1,5 @@
 import random
+import warnings
 
 import torch
 
@@ -12,10 +13,11 @@ def _transform(dat, irreps_dat, rot_mat, translation=0.):
         if irreps is None:
             out.append(a)
         elif irreps == 'cartesian_points':
-            out.append((a @ rot_mat.T) + translation)
+            translation = torch.as_tensor(translation, device=a.device)
+            out.append((a @ rot_mat.T.to(a.device)) + translation)
         else:
             # For o3.Irreps
-            out.append(a @ irreps.D_from_matrix(rot_mat).T)
+            out.append(a @ irreps.D_from_matrix(rot_mat).T.to(a.device))
     return out
 
 
@@ -23,31 +25,44 @@ def _get_io_irreps(func, irreps_in=None, irreps_out=None):
     """Preprocess or, if not given, try to infer the I/O irreps for ``func``."""
     SPECIAL_VALS = ['cartesian_points', None]
 
+    if (irreps_in is None or irreps_out is None) and isinstance(func, torch.jit.ScriptModule):
+        warnings.warn(
+            "Asking to infer irreps in/out of a compiled TorchScript module. This is unreliable, please provide `irreps_in` and `irreps_out` explicitly."
+        )
+
     if irreps_in is None:
         if hasattr(func, 'irreps_in'):
-            irreps_in = [func.irreps_in]
+            irreps_in = func.irreps_in  # gets checked for type later
         elif hasattr(func, 'irreps_in1'):
             irreps_in = [func.irreps_in1, func.irreps_in2]
         else:
             raise ValueError("Cannot infer irreps_in for %r; provide them explicitly" % func)
     if irreps_out is None:
         if hasattr(func, 'irreps_out'):
-            irreps_out = [func.irreps_out]
+            irreps_out = func.irreps_out  # gets checked for type later
         else:
             raise ValueError("Cannot infer irreps_out for %r; provide them explicitly" % func)
 
     if isinstance(irreps_in, Irreps) or irreps_in in SPECIAL_VALS:
         irreps_in = [irreps_in]
-    elif isinstance(irreps_in, list) or isinstance(irreps_in, tuple):
+    elif isinstance(irreps_in, list):
         irreps_in = [i if i in SPECIAL_VALS else Irreps(i) for i in irreps_in]
     else:
+        if isinstance(irreps_in, tuple) and not isinstance(irreps_in, Irreps):
+            warnings.warn(
+                f"Module {func} had irreps_in of type tuple but not Irreps; ambiguous whether the tuple should be interpreted as a tuple representing a single Irreps or a tuple of objects each to be converted to Irreps. Assuming the former. If the latter, use a list."
+            )
         irreps_in = [Irreps(irreps_in)]
 
     if isinstance(irreps_out, Irreps) or irreps_out in SPECIAL_VALS:
         irreps_out = [irreps_out]
-    elif isinstance(irreps_out, list) or isinstance(irreps_out, tuple):
+    elif isinstance(irreps_out, list):
         irreps_out = [i if i in SPECIAL_VALS else Irreps(i) for i in irreps_out]
     else:
+        if isinstance(irreps_in, tuple) and not isinstance(irreps_in, Irreps):
+            warnings.warn(
+                f"Module {func} had irreps_out of type tuple but not Irreps; ambiguous whether the tuple should be interpreted as a tuple representing a single Irreps or a tuple of objects each to be converted to Irreps. Assuming the former. If the latter, use a list."
+            )
         irreps_out = [Irreps(irreps_out)]
 
     return irreps_in, irreps_out
@@ -82,14 +97,47 @@ def _get_device(mod: torch.nn.Module) -> torch.device:
     return a_buf.device if a_buf is not None else 'cpu'
 
 
-def _to_device(args, device):
+_FLOATING_DTYPES = [torch.float32, torch.float64]
+
+
+def _get_floating_dtype(mod: torch.nn.Module) -> torch.dtype:
+    """Guess floating dtype for module.
+
+    Assumes no mixed precision.
+    """
+    # Try to a get a parameter
+    a_buf = None
+    for buf in mod.parameters():
+        if buf.dtype in _FLOATING_DTYPES:
+            a_buf = buf
+            break
+    if a_buf is None:
+        # If there isn't one, try to get a buffer
+        for buf in mod.buffers():
+            if buf.dtype in _FLOATING_DTYPES:
+                a_buf = buf
+                break
+    return a_buf.dtype if a_buf is not None else torch.get_default_dtype()
+
+
+def _to_device_dtype(args, device=None, dtype=None):
+    kwargs = {}
+    if device is not None:
+        kwargs['device'] = device
+    if dtype is not None:
+        kwargs['dtype'] = dtype
+
     if isinstance(args, torch.Tensor):
-        return args.to(device=device)
+        if args.dtype in _FLOATING_DTYPES:
+            # Only convert dtypes of floating tensors
+            return args.to(device=device, dtype=dtype)
+        else:
+            return args.to(device=device)
     elif isinstance(args, tuple):
-        return tuple(_to_device(e, device) for e in args)
+        return tuple(_to_device_dtype(e, **kwargs) for e in args)
     elif isinstance(args, list):
-        return [_to_device(e, device) for e in args]
+        return [_to_device_dtype(e, **kwargs) for e in args]
     elif isinstance(args, dict):
-        return{k: _to_device(v, device) for k, v in args.items()}
+        return{k: _to_device_dtype(v, **kwargs) for k, v in args.items()}
     else:
-        raise TypeError("Only (nested) dict/tuple/lists of Tensors can be moved to a device.")
+        raise TypeError("Only (nested) dict/tuple/lists of Tensors can be moved to a device/dtype.")
